/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package main

import (
	"context"
	"fmt"
	"html/template"
	"log"
	"net"
	"net/http"
	"os"
	"path"
	"regexp"
	"sort"
	"strings"
	"time"
)

var goGetTmpl *template.Template

func init() {
	const respTmpl = `<!doctype html>` +
		`<html>` +
		`<head>` +
		`<meta name="go-import"` +
		` content="{{.Module}} git {{.ImportURI}}">` +
		`<meta name="go-source"` +
		` content="{{.Module}}` +
		` {{print .SourceURI}}` +
		` {{.SourceURI}}/src/{{.Branch}}{/dir}` +
		` {{.SourceURI}}/src/{{.Branch}}{/dir}/{file}#L{line}">` +
		`</head>` +
		`<body>` +
		`go get https://{{.Module}}` +
		`</body>` +
		`</html>`

	goGetTmpl =
		template.Must(template.New("resp").Parse(respTmpl))
}

func NewServer(ctx context.Context) *Server {
	accessToken, ok := os.LookupEnv("GOGS_ACCESS_TOKEN")
	if !ok {
		panic("Missing environment variable GOGS_ACCESS_TOKEN.")
	}
	return &Server{
		ctx:    ctx,
		client: NewGogsClient(ctx, accessToken),
	}
}

type Server struct {
	ctx      context.Context
	svr      *http.Server
	client   *GogsClient
	projects Cached[map[string]GitRepo]
}

func (s *Server) ListenAndServe(port int) error {
	s.svr = &http.Server{
		Addr:         fmt.Sprintf(":%d", port),
		ReadTimeout:  10 * time.Second,
		WriteTimeout: 10 * time.Second,
		BaseContext: func(_ net.Listener) context.Context {
			return s.ctx
		},
	}
	s.paths()
	return s.svr.ListenAndServe()
}

func (s *Server) Shutdown() error {
	return s.svr.Shutdown(s.ctx)
}

func (s *Server) paths() {
	http.HandleFunc("/go/", s.handleGoGet())
}

func (s *Server) updateProjectCache(group string) error {
	if s.projects.Expired() {
		projects := map[string]GitRepo{}
		pager := FetchOrgRepos(s.client, group)
		for p := range pager.Iter() {
			projects[p.Path()] = p
		}
		if err := pager.LastErr(); err != nil {
			log.Printf("request go-get: unable to fetch gogs projects: %v", err)
			return err
		}
		s.projects = NewCached(&projects, 30*time.Minute)
	}
	return nil
}

func (s *Server) handleGoGet() http.HandlerFunc {
	major := regexp.MustCompile(`^v[0-9]+$`)
	goSubpath := regexp.MustCompile(`/go/`)

	return func(w http.ResponseWriter, req *http.Request) {
		clean := path.Clean(req.URL.Path)
		group := req.Host

		err := s.updateProjectCache(group)
		if err != nil {
			code := http.StatusServiceUnavailable
			http.Error(w, http.StatusText(code), code)
			return
		}

		// /@v/$version.{.info,.mod,.zip}
		version := ""
		ext := ""
		if ext = path.Ext(clean); ext != "" {
			version = strings.TrimSuffix(path.Base(clean), ext)
			// TODO Dispatch to .info, .mod, or .zip handler
			log.Printf("request %s.%s", version, ext)
			code := http.StatusNotFound
			http.Error(w, http.StatusText(code), code)
		}

		// /@v/list
		repo := ""
		head, tail := path.Split(clean)
		if tail == "list" {
			if atV := path.Base(head); atV == "@v" {
				repo = path.Base(path.Dir(path.Dir(head)))
				tags := make([]*GogsTag, 0, 16)
				pager := ListTags(s.client, group, repo)
				for r := range pager.Iter() {
					tags = append(tags, r)
				}
				sort.SliceStable(tags, func(i, j int) bool {
					ti := tags[i].Name
					tj := tags[j].Name
					return strings.Compare(ti, tj) == -1
				})
				for _, t := range tags {
					w.Write([]byte(t.Name + "\n"))
				}
				if pager.LastErr() != nil {
					log.Printf("request /@v/list: list gogs releases for %s/%s: %v", group, repo, pager.LastErr())
					code := http.StatusServiceUnavailable
					http.Error(w, http.StatusText(code), code)
				}
				return
			}
			log.Printf("malformed request for /@v/list: %s", req.URL.String())
			code := http.StatusBadRequest
			http.Error(w, http.StatusText(code), code)
			return
		}

		module := path.Join(group, head)
		if major.MatchString(tail) {
			repo = path.Base(head)
		} else {
			module = path.Join(module, tail)
			repo = tail
		}
		// example.com/go/blarg/v2 |-> example.com/go/blarg
		// example.com/go/blarg |-> example.com/go, blarg
		modPath := goSubpath.ReplaceAllString(module, "/")

		project, ok := (*s.projects.Value())[modPath]
		if !ok {
			code := http.StatusNotFound
			http.Error(w, http.StatusText(code), code)
			log.Printf("request go-get: unable to find git project for module '%s'", module)
			return
		}

		if q := req.URL.Query(); q.Get("go-get") == "1" {
			// TODO this may write a partial result on error; don't
			err = goGetTmpl.Execute(w, struct {
				Module    string
				ImportURI string
				SourceURI string
				Branch    string
			}{module,
				project.GitURI(),
				project.WebURI(),
				project.DefaultBranch(),
			})
			if err != nil {
				code := http.StatusInternalServerError
				http.Error(w, http.StatusText(code), code)
				log.Printf("request go-get: execute template: '%s': %v", modPath, err)
			}
			return
		}
	}
}
