/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package main

import (
	"bytes"
	"context"
	"fmt"
	"net/http"
	"os"
	"strings"
	"testing"
)

var srv *Server

func TestMain(m *testing.M) {
	ctx := context.Background()
	srv = NewServer(ctx)

	ret := m.Run()
	os.Exit(ret)
}

func TestGoProxyProtocol(t *testing.T) {
	t.Run(
		"Receive expected HTML on go get request for existing module on Gogs",
		func(t *testing.T) {
			module := "idio.link/go/netaddr"
			expected := `<!doctype html>` +
				`<html>` +
				`<head>` +
				`<meta name="go-import"` +
				` content="idio.link/go/netaddr git https://git.idiolink.net/idio.link/netaddr.git">` +
				`<meta name="go-source"` +
				` content="idio.link/go/netaddr` +
				` https://git.idiolink.net/idio.link/netaddr` +
				` https://git.idiolink.net/idio.link/netaddr/src/master{/dir}` +
				` https://git.idiolink.net/idio.link/netaddr/src/master{/dir}/{file}#L{line}` +
				`">` +
				`</head>` +
				`<body>` +
				`go get https://idio.link/go/netaddr` +
				`</body>` +
				`</html>`

			reqURI := fmt.Sprintf("https://%s/v2?go-get=1", module)
			req, _ := http.NewRequest("GET", reqURI, nil)
			w := NewFakeWriter()
			srv.handleGoGet()(w, req)

			if strings.Compare(w.buf.String(), expected) != 0 {
				t.Errorf("expected:\n%s\nactual:\n%s", expected, w.buf.String())
			}
		},
	)
	t.Run(
		"Receive list of module versions for existing module on Gogs",
		func(t *testing.T) {
			module := "idio.link/go/netaddr"
			expected := `v0.0.1
v1.0.0
v1.1.0
v1.2.0
v1.2.1
v1.3.0
v1.4.0
v1.5.0
v1.6.0
v2.0.0
v2.0.1
v2.1.0
v2.1.1
`

			reqURI := fmt.Sprintf("https://%s/@v/list", module)
			req, _ := http.NewRequest("GET", reqURI, nil)
			w := NewFakeWriter()
			srv.handleGoGet()(w, req)

			if strings.Compare(w.buf.String(), expected) != 0 {
				t.Errorf("expected:\n%s\nactual:\n%s", expected, w.buf.String())
			}
		},
	)
}

func NewFakeWriter() *FakeWriter {
	buf := make([]byte, 0, 2048)

	return &FakeWriter{
		buf: bytes.NewBuffer(buf),
	}
}

type FakeWriter struct {
	buf    *bytes.Buffer
	status int
}

func (w *FakeWriter) Header() http.Header {
	h := make(map[string][]string)

	return h
}

func (w *FakeWriter) Write(b []byte) (int, error) {
	return w.buf.Write(b)
}

func (w *FakeWriter) WriteHeader(status int) {
	w.status = status
}
